/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>

#include <lcms2.h>

#include <bisection.h>
#include <nmsimplex.h>
#include <dngref.h>
#include <interp.h>
#include <look.h>

struct look_tone_rep_op_t_ {
    cmsHANDLE h02;
    v3 wp;
    const double *tc;
    int tc_len;
    double *itc;
    int itc_len;
    struct tone_rep_op_config conf;
};

struct find_tc_slope_fun_arg {
    const double *tc;
    int tc_len;
};

static double
find_tc_slope_fun(double k,
                  void *arg)
{
    struct find_tc_slope_fun_arg *a = (struct find_tc_slope_fun_arg *)arg;
    double areasum = 0;
    const int steps = 256;
    for (int i = 0; i < steps; i++) {
        double x = 0.1 + ((double)i / (steps-1)) * 0.5; // testing (sRGB) range [0.1 - 0.6], ie ignore highligths and dark shadows
        double y = srgb_gamma_forward(curve_y(srgb_gamma_inverse(x), a->tc, a->tc_len));
        double y1 = k * x;
        if (y1 > 1) y1 = 1;
        areasum += (y - y1) * (y - y1); // square is a rough approx of (twice) the area, but it's fine for our purposes
    }
    return areasum;
}

// calculate a single value that represents the contrast of the tone curve
static double
calculate_tonecurve_contrast_value(const double tc[],
                                   int tc_len)
{

    // find linear y = k*x the best approximates the curve, which is the linear scaling/exposure component that does not contribute any contrast

    // Note: the analysis is made on the gamma encoded curve, as the LUT is linear we make backwards gamma to
    struct find_tc_slope_fun_arg arg = { tc, tc_len };
    double k = interval_halving_min(find_tc_slope_fun, &arg, 0.1, 5.0, 0.001, 10000);
    //fprintf(stderr, "average slope: %f\n", k);

    double maxslope = 0;
    {
        // look at midtone slope
        const double xd = 0.07;
        const double tx[] = { 0.30, 0.35, 0.40, 0.45 }; // we only look in the midtone range
        for (int i = 0; i < sizeof(tx)/sizeof(tx[0]); i++) {
            double x0 = tx[i] - xd;
            double y0 = srgb_gamma_forward(curve_y(srgb_gamma_inverse(x0), tc, tc_len)) - k * x0;
            double x1 = tx[i] + xd;
            double y1 = srgb_gamma_forward(curve_y(srgb_gamma_inverse(x1), tc, tc_len)) - k * x1;
            double slope = 1.0 + (y1 - y0) / (x1 - x0);
            if (slope > maxslope) {
                maxslope = slope;
            }
        }

        // look at slope at (light) shadows and (dark) highlights
        double e_maxslope = 0;
        {
            const double tx[] = { 0.20, 0.25, 0.50, 0.55 }; // we only look in the midtone range
            for (int i = 0; i < sizeof(tx)/sizeof(tx[0]); i++) {
                double x0 = tx[i] - xd;
                double y0 = srgb_gamma_forward(curve_y(srgb_gamma_inverse(x0), tc, tc_len)) - k * x0;
                double x1 = tx[i] + xd;
                double y1 = srgb_gamma_forward(curve_y(srgb_gamma_inverse(x1), tc, tc_len)) - k * x1;
                double slope = 1.0 + (y1 - y0) / (x1 - x0);
                if (slope > e_maxslope) {
                    e_maxslope = slope;
                }
            }
        }
        //fprintf(stderr, "%.3f %.3f\n", maxslope, e_maxslope);
        // midtone slope is more important for contrast, but weigh in some slope from brights and darks too.
        maxslope = maxslope * 0.7 + e_maxslope * 0.3;
    }
    return maxslope;
}

void
look_neutral_tone_rep_op_tc_analysis(const double tc[],
                                     int tc_len,
                                     double *cv,
                                     double *cs)
{
    const double p[] = {
        0.60, 0.70, // lowest contrast
        0.70, 0.80,
        0.90, 0.94,
        0.99, 1.00,
        1.00, 1.00, // 1.0 (linear curve) to 1.0, no scaling
        1.07, 1.00,
        1.08, 1.00,
        1.11, 1.02,
        1.20, 1.08,
        1.30, 1.12,
        1.80, 1.20,
        2.00, 1.22  // highest contrast
    };

    double cf[1000];
    double cf_range[2];
    const size_t in_len = sizeof(p)/sizeof(p[0])/2;
    double in_x[in_len];
    double in_y[in_len];
    for (size_t i = 0; i < in_len; i++) {
        in_x[i]= p[2*i+0];
        in_y[i]= p[2*i+1];
    }
    const size_t out_len = sizeof(cf)/sizeof(cf[0]);
    double out_x[out_len];
    for (size_t i = 0; i < out_len; i++) {
        out_x[i] = in_x[0] + (in_x[in_len-1] - in_x[0]) * (double)i / (out_len-1);
    }
    cubic_spline(in_x, in_y, in_len, out_x, cf, out_len);
    cf_range[0] = in_x[0];
    cf_range[1] = in_x[in_len-1];

    *cv = calculate_tonecurve_contrast_value(tc, tc_len);
    *cs = curve_y_rng(*cv, cf_range, cf, sizeof(cf)/sizeof(cf[0]));
}

look_tone_rep_op_t *
look_tone_rep_op_new(const double tc[],
                     int tc_len,
                     const struct tone_rep_op_config *conf)
{
    look_tone_rep_op_t *state = calloc(1, sizeof(*state));
    cmsViewingConditions vc;
    vc.whitePoint = *cmsD50_XYZ();
    vc.whitePoint.X *= 100;
    vc.whitePoint.Y *= 100;
    vc.whitePoint.Z *= 100;
    vc.Yb = 20;
    vc.La = 20;
    vc.surround = AVG_SURROUND;
    vc.D_value = 1.0;
    state->h02 = cmsCIECAM02Init(0, &vc);
    state->wp = (v3){{ vc.whitePoint.X/100, vc.whitePoint.Y/100, vc.whitePoint.Z/100 }};

    if (conf != NULL) {
        state->conf = *conf;
    } else {
        state->conf.chroma_scaling = -1;

        state->conf.saturated.adjust_factor = 0.95;
        state->conf.saturated.low_chroma = 35;
        state->conf.saturated.high_chroma = 60;

        state->conf.shadows.adjust_factor = 1.20;
        state->conf.shadows.low_lightness = 0.15;
        state->conf.shadows.high_lightness = 0.50;

        state->conf.rolloff.keep_factor = calloc(1, sizeof(struct lookop_curve) + sizeof(double) * 4);
        state->conf.rolloff.keep_factor->handle_count = 2;
        state->conf.rolloff.keep_factor->handles[0] = 0;
        state->conf.rolloff.keep_factor->handles[1] = 0.20;
        state->conf.rolloff.keep_factor->handles[2] = 360;
        state->conf.rolloff.keep_factor->handles[3] = 0.20;
        state->conf.rolloff.low_satscale = calloc(1, sizeof(struct lookop_curve) + sizeof(double) * 4);
        state->conf.rolloff.low_satscale->type = LOOKOP_CURVE_LINEAR;
        state->conf.rolloff.low_satscale->handle_count = 2;
        state->conf.rolloff.low_satscale->handles[0] = 0;
        state->conf.rolloff.low_satscale->handles[1] = 1.00;
        state->conf.rolloff.low_satscale->handles[2] = 360;
        state->conf.rolloff.low_satscale->handles[3] = 1.00;
        state->conf.rolloff.high_satscale = calloc(1, sizeof(struct lookop_curve) + sizeof(double) * 4);
        state->conf.rolloff.high_satscale->handle_count = 2;
        state->conf.rolloff.high_satscale->handles[0] = 0;
        state->conf.rolloff.high_satscale->handles[1] = 1.20;
        state->conf.rolloff.high_satscale->handles[2] = 360;
        state->conf.rolloff.high_satscale->handles[3] = 1.20;
    }

    if (state->conf.chroma_scaling < 0) {
        double cv, cs;
        look_neutral_tone_rep_op_tc_analysis(tc, tc_len, &cv, &cs);
        state->conf.chroma_scaling = cs;
    }

    state->tc = tc;
    state->tc_len = tc_len;

    return state;
}

void
look_tone_rep_op_delete(look_tone_rep_op_t *state)
{
    if (state == NULL) {
        return;
    }
    cmsCIECAM02Done(state->h02);
    free(state);
}

static double
lookop_curve(const struct lookop_curve *c,
             double x)
{
    double in_x[c->handle_count];
    double in_y[c->handle_count];
    for (int i = 0; i < c->handle_count; i++) {
        in_x[i]= c->handles[2*i+0];
        in_y[i]= c->handles[2*i+1];
    }
    double y;
    switch (c->type) {
    case LOOKOP_CURVE_SPLINE:
        cubic_spline(in_x, in_y, c->handle_count, &x, &y, 1);
        break;
    case LOOKOP_CURVE_ROUNDEDSTEP:
        rounded_linear_interpolation(in_x, in_y, c->handle_count, &x, &y, 1);
        break;
    case LOOKOP_CURVE_LINEAR:
        linear_interpolation(in_x, in_y, c->handle_count, &x, &y, 1);
        break;
    }
    return y;
}

static double
get_vs(v3 jch,
       v3 rgb,
       int vs_count,
       const struct lookop_vs *vs_array)
{
    double s = 1.0;

    for (int i = 0; i < vs_count; i++) {
        const struct lookop_vs *vs = &vs_array[i];
        double x = 0;
        double diff = vs->xrange[1] - vs->xrange[0];
        double start = vs->xrange[0];
        switch (vs->x_type) {
        case LOOKOP_VS_X_LIGHTNESS: x = jch.v[0]; break;
        case LOOKOP_VS_X_CHROMA: x = jch.v[1]; break;
        case LOOKOP_VS_X_HUE:
            x = jch.v[2];
            if (start < 0) start += 360;
            if (start + diff > 360 && x < vs->xrange[1]) x += 360;
            break;
        case LOOKOP_VS_X_HSL_LIGHTNESS:
        case LOOKOP_VS_X_HSL_SATURATION:
        case LOOKOP_VS_X_HSL_HUE:
        case LOOKOP_VS_X_HSV_VALUE:
        case LOOKOP_VS_X_HSV_SATURATION:
        case LOOKOP_VS_X_HSV_HUE: {
            v3 hsl = rgb2hsl(rgb);
            v3 hsv = dcp_rgb2hsv(rgb);
            hsv.v[0] /= 6.0;
            switch (vs->x_type) {
            default: abort();
            case LOOKOP_VS_X_HSL_LIGHTNESS: x = 100 * hsl.v[2]; break;
            case LOOKOP_VS_X_HSL_SATURATION: x = 100 * hsl.v[1]; break;
            case LOOKOP_VS_X_HSL_HUE:
                x = hsl.v[0] * 360.0;
                if (start < 0) start += 360;
                if (start + diff > 360 && x < vs->xrange[1]) x += 360;
                break;
            case LOOKOP_VS_X_HSV_VALUE: x = 100 * hsv.v[2]; break;
            case LOOKOP_VS_X_HSV_SATURATION: x = 100 * hsv.v[1]; break;
            case LOOKOP_VS_X_HSV_HUE:
                x = hsv.v[0] * 360.0;
                if (start < 0) start += 360;
                if (start + diff > 360 && x < vs->xrange[1]) x += 360;
                break;
            }
            break;
        }
        }
        x -= start;
        x /= diff;
        if (x < 0) x = 0;
        if (x > 1) x = 1;
        if (!isfinite(x)) x = 0;
        double y = lookop_curve(vs->curve, x);
        y *= vs->yscale;
        s *= y;
    }

    return s;
}

static v3
apply_look_operators(v3 jch,
                     const m3x3 rgb2xyz,
                     const m3x3 xyz2rgb,
                     look_tone_rep_op_t *state)
{
    if (state->conf.lookop_count <= 0) {
        return jch;
    }

    for (int i = 0; i < state->conf.lookop_count; i++) {
        const struct lookop *lo = &state->conf.lookops[i];

        v3 rgb, xyz;
        {
            cmsCIEXYZ XYZ;
            cmsJCh JCh;
            JCh = (cmsJCh){ jch.v[0], jch.v[1], jch.v[2] };
            cmsCIECAM02Reverse(state->h02, &JCh, &XYZ);
            xyz = (v3){{ XYZ.X / 100.0, XYZ.Y / 100.0, XYZ.Z / 100.0 }};
            for (int i = 0; i < 3; i++) {
                if (!isfinite(xyz.v[i])) {
                    return jch;
                }
            }
            rgb = m3x3_mul_v3(xyz2rgb, xyz);
            for (int i = 0; i < 3; i++) {
                if (!isfinite(rgb.v[i])) return jch;
                if (rgb.v[i] < 0) rgb.v[i] = 0;
                if (rgb.v[i] > 1) rgb.v[i] = 1;
            }
        }
        double vs = get_vs(jch, rgb, lo->vs_count, (const struct lookop_vs *)lo->vs);
        if (lo->blend_invert) vs = 1.0 - vs;

        v3 new_jch = jch;
        switch (lo->type) {
        case LOOKOP_STRETCH:
            for (int j = 0; j < lo->p.stretch.dim_count; j++) {
                double x = 0;
                switch (lo->p.stretch.dims[j]) {
                case LOOKOP_VS_X_LIGHTNESS: x = jch.v[0]; break;
                case LOOKOP_VS_X_CHROMA: x = jch.v[1]; break;
                case LOOKOP_VS_X_HUE: x = jch.v[2]; break;
                default: abort();
                }
                x -= lo->p.stretch.ranges[j][0];
                x /= lo->p.stretch.ranges[j][1] - lo->p.stretch.ranges[j][0];
                if (x < 0 || x > 1 || !isfinite(x)) {
                    continue;
                }
                double y = lookop_curve(lo->p.stretch.curves[j], x);
                y *= lo->p.stretch.ranges[j][1] - lo->p.stretch.ranges[j][0];
                y += lo->p.stretch.ranges[j][0];
                switch (lo->p.stretch.dims[j]) {
                case LOOKOP_VS_X_LIGHTNESS: new_jch.v[0] = y; break;
                case LOOKOP_VS_X_CHROMA: new_jch.v[1] = y; break;
                case LOOKOP_VS_X_HUE: new_jch.v[2] = y; break;
                default: abort();
                }
            }
            break;
        case LOOKOP_ADDHUE:
            new_jch.v[2] += lo->p.gen.value;
            if (new_jch.v[2] > 360) new_jch.v[2] -= 360;
            if (new_jch.v[2] < 0)   new_jch.v[2] += 360;
            break;
        case LOOKOP_ADDCHROMA:
            new_jch.v[1] += lo->p.gen.value;
            break;
        case LOOKOP_ADDLIGHTNESS:
            new_jch.v[0] += lo->p.gen.value;
            break;
        case LOOKOP_SCALECHROMA:
            new_jch.v[1] *= lo->p.gen.value;
            break;
        case LOOKOP_SCALELIGHTNESS:
            new_jch.v[0] *= lo->p.gen.value;
            break;
        case LOOKOP_SETTEMPERATURE: {
            cmsCIEXYZ XYZ;
            cmsJCh JCh;
            v3 wp = dcp_d50();
            wp = xyY2xyz((v3){{ lo->p.settemp.xy[0], lo->p.settemp.xy[1], wp.v[1] }});
            v3 new_xyz = chromatic_adaptation_transform(CAT_CAT02, xyz, wp, dcp_d50());
            XYZ = (cmsCIEXYZ){ .X = new_xyz.v[0] * 100.0, .Y = new_xyz.v[1] * 100.0, .Z = new_xyz.v[2] * 100.0 };
            cmsCIECAM02Forward(state->h02, &XYZ, &JCh);
            if (!isfinite(JCh.J) || !isfinite(JCh.C) || !isfinite(JCh.h)) {
                return jch;
            }
            new_jch = (v3){{ JCh.J, JCh.C, JCh.h }};
            break;
        }
        case LOOKOP_CURVES: {
            cmsCIEXYZ XYZ;
            cmsJCh JCh;
            v3 new_rgb;
            for (int i = 0; i < 3; i++) {
                double x = rgb.v[i];
                if (lo->p.curves.is_srgb_gamma) {
                    x = srgb_gamma_forward(x);
                } else if (lo->p.curves.gamma != 1.0) {
                    x = pow(x, 1.0 / lo->p.curves.gamma);
                }
                double y = lookop_curve(lo->p.curves.curves[i], x);
                if (lo->p.curves.is_srgb_gamma) {
                    y = srgb_gamma_inverse(y);
                } else if (lo->p.curves.gamma != 1.0) {
                    y = pow(y, lo->p.curves.gamma);
                }
                new_rgb.v[i] = y;
            }
            v3 new_xyz = m3x3_mul_v3(rgb2xyz, new_rgb);
            XYZ = (cmsCIEXYZ){ .X = new_xyz.v[0] * 100.0, .Y = new_xyz.v[1] * 100.0, .Z = new_xyz.v[2] * 100.0 };
            cmsCIECAM02Forward(state->h02, &XYZ, &JCh);
            if (!isfinite(JCh.J) || !isfinite(JCh.C) || !isfinite(JCh.h)) {
                return jch;
            }
            new_jch = (v3){{ JCh.J, JCh.C, JCh.h }};
            break;
        }
        }

        if (lo->blend_rgb) {
            cmsCIEXYZ XYZ;
            cmsJCh JCh;

            v3 new_rgb;
            {
                JCh = (cmsJCh){ new_jch.v[0], new_jch.v[1], new_jch.v[2] };
                cmsCIECAM02Reverse(state->h02, &JCh, &XYZ);
                v3 xyz1 = (v3){{ XYZ.X / 100.0, XYZ.Y / 100.0, XYZ.Z / 100.0 }};
                v3 rgb1 = m3x3_mul_v3(xyz2rgb, xyz1);
                for (int i = 0; i < 3; i++) {
                    if (!isfinite(rgb1.v[i])) return jch;
                    if (rgb1.v[i] < 0) rgb1.v[i] = 0;
                    if (rgb1.v[i] > 1) rgb1.v[i] = 1;
                }
                new_rgb = rgb1;
            }

            rgb.v[0] = srgb_gamma_forward(rgb.v[0]) * (1 - vs) + srgb_gamma_forward(new_rgb.v[0]) * vs;
            rgb.v[1] = srgb_gamma_forward(rgb.v[1]) * (1 - vs) + srgb_gamma_forward(new_rgb.v[1]) * vs;
            rgb.v[2] = srgb_gamma_forward(rgb.v[2]) * (1 - vs) + srgb_gamma_forward(new_rgb.v[2]) * vs;
            for (int i = 0; i < 3; i++) {
                if (!isfinite(rgb.v[i])) return jch;
                if (rgb.v[i] < 0) rgb.v[i] = 0;
                if (rgb.v[i] > 1) rgb.v[i] = 1;
                rgb.v[i] = srgb_gamma_inverse(rgb.v[i]);
            }
            xyz = m3x3_mul_v3(rgb2xyz, rgb);
            XYZ = (cmsCIEXYZ){ .X = xyz.v[0] * 100.0, .Y = xyz.v[1] * 100.0, .Z = xyz.v[2] * 100.0 };
            cmsCIECAM02Forward(state->h02, &XYZ, &JCh);
            if (!isfinite(JCh.J) || !isfinite(JCh.C) || !isfinite(JCh.h)) {
                return jch;
            }
            jch = (v3){{ JCh.J, JCh.C, JCh.h }};
        } else {
            jch = jch_blend(new_jch, jch, vs);
        }

    }
    return jch;
}

bool
look_tone_rep_op(v3 *out,
                 const m3x3 rgb2xyz,
                 const m3x3 xyz2rgb,
                 v3 xyz, // D50!
                 look_tone_rep_op_t *state)
{
    const struct tone_rep_op_config *conf = &state->conf;
    cmsCIEXYZ XYZ;
    cmsJCh JCh;

    if (xyz.v[0] == 0.0 && xyz.v[1] == 0.0 && xyz.v[2] == 0.0) {
        *out = xyz;
        return true;
    }

    v3 orig_rgb = m3x3_mul_v3(xyz2rgb, xyz);;
    for (int i = 0; i < 3; i++) {
        if (orig_rgb.v[i] < 0) orig_rgb.v[i] = 0;
        if (orig_rgb.v[i] > 1) orig_rgb.v[i] = 1;
    }

    v3 argb;
    {
        argb = dcp_apply_tonecurve(orig_rgb, state->tc, state->tc_len);
	if (argb.v[0] >= 1 && argb.v[1] >= 1 && argb.v[2] >= 1) {
            // clip fast path, will also avoid strange colors of clipped highlights
            *out = m3x3_mul_v3(rgb2xyz, (v3){{1,1,1}});
            return true;
	}
    }
    const double Yr = rgb2xyz.v[1][0];
    const double Yg = rgb2xyz.v[1][1];
    const double Yb = rgb2xyz.v[1][2];

    double oldLuminance = orig_rgb.v[0]*Yr + orig_rgb.v[1]*Yg + orig_rgb.v[2]*Yb;
    double newLuminance = argb.v[0]*Yr + argb.v[1]*Yg + argb.v[2]*Yb;
    double Lcoef = newLuminance/oldLuminance;
    v3 rgb = {{ orig_rgb.v[0] * Lcoef, orig_rgb.v[1] * Lcoef, orig_rgb.v[2] * Lcoef }};

    xyz = m3x3_mul_v3(rgb2xyz, rgb);
    XYZ = (cmsCIEXYZ){ .X = xyz.v[0] * 100.0, .Y = xyz.v[1] * 100.0, .Z = xyz.v[2] * 100.0 };
    cmsCIECAM02Forward(state->h02, &XYZ, &JCh);
    if (!isfinite(JCh.J) || !isfinite(JCh.C) || !isfinite(JCh.h)) {
        return false;
    }

    double cmul = conf->chroma_scaling;

    // depending on color, the chroma scaling factor can be fine-tuned below

    if (conf->saturated.adjust_factor != 1.0 && conf->saturated.adjust_factor >= 0.0) {
        // decrease chroma scaling sligthly of extremely saturated colors
        double saturated_scale_factor = conf->saturated.adjust_factor;
        const double lolim = conf->saturated.low_chroma;
        const double hilim = conf->saturated.high_chroma;
        if (JCh.C < lolim) {
            // chroma is low enough, don't scale
            saturated_scale_factor = 1.0;
        } else if (JCh.C < hilim) {
            // S-curve transition between low and high limit
            double x = (JCh.C - lolim) / (hilim - lolim); // x = [0..1], 0 at lolim, 1 at hilim
            if (x < 0.5) {
                x = 0.5 * pow(2*x, 2);
            } else {
                x = 0.5 + 0.5 * (1-pow(1-2*(x-0.5), 2));
            }
            saturated_scale_factor = 1.0*(1.0-x) + saturated_scale_factor*x;
        } else {
            // do nothing, high saturation color, keep scale factor
        }
        cmul *= saturated_scale_factor;
    }

    if (conf->shadows.adjust_factor != 1.0 && conf->shadows.adjust_factor >= 0.0) {
        // increase chroma scaling slightly of shadows
        double nL = srgb_gamma_forward(newLuminance); // apply gamma so we make comparison and transition with a more perceptual lightness scale
        double dark_scale_factor = conf->shadows.adjust_factor;
        //double dark_scale_factor = 1.0 + state.debug.p2 / 100.0f;
        const double lolim = conf->shadows.low_lightness;
        const double hilim = conf->shadows.high_lightness;
        if (nL < lolim) {
            // do nothing, keep scale factor
        } else if (nL < hilim) {
            // S-curve transition
            double x = (nL - lolim) / (hilim - lolim); // x = [0..1], 0 at lolim, 1 at hilim
            if (x < 0.5) {
                x = 0.5 * pow(2*x, 2);
            } else {
                x = 0.5 + 0.5 * (1-pow(1-2*(x-0.5), 2));
            }
            dark_scale_factor = dark_scale_factor*(1.0-x) + 1.0*x;
        } else {
            dark_scale_factor = 1.0;
        }
        cmul *= dark_scale_factor;
    }

    { // to avoid strange CIECAM02 chroma errors on close-to-shadow-clipping colors we reduce chroma scaling towards 1.0 for very-close-to-black
        double dark_scale_factor = 1.0 / cmul;
        const double lolim = 4;
        const double hilim = 7;
        if (JCh.J < lolim) {
            // do nothing, keep scale factor
        } else if (JCh.J < hilim) {
            // S-curve transition
            double x = (JCh.J - lolim) / (hilim - lolim);
            if (x < 0.5) {
                x = 0.5 * pow(2*x, 2);
            } else {
                x = 0.5 + 0.5 * (1-pow(1-2*(x-0.5), 2));
            }
            dark_scale_factor = dark_scale_factor*(1.0-x) + 1.0*x;
        } else {
            dark_scale_factor = 1.0;
        }
        cmul *= dark_scale_factor;
    }

    v3 jch = desaturate_to_fit((v3){{ JCh.J, JCh.C * cmul, JCh.h }}, xyz2rgb, state->h02);
    JCh = (cmsJCh){ jch.v[0], jch.v[1], jch.v[2] };
    cmsCIECAM02Reverse(state->h02, &JCh, &XYZ);
    xyz = (v3){{ XYZ.X / 100.0, XYZ.Y / 100.0, XYZ.Z / 100.0 }};
    rgb = m3x3_mul_v3(xyz2rgb, xyz);
    for (int i = 0; i < 3; i++) {
        if (!isfinite(rgb.v[i])) return false;
        if (rgb.v[i] < 0) rgb.v[i] = 0;
        if (rgb.v[i] > 1) rgb.v[i] = 1;
    }

    {
        v3 ahsv = dcp_rgb2hsv(argb);
        v3 hsv = dcp_rgb2hsv(rgb);
        double keep = lookop_curve(conf->rolloff.keep_factor, hsv.v[0] / 6.0 * 360);
        const double lolim = lookop_curve(conf->rolloff.low_satscale, hsv.v[0] / 6.0 * 360);
        const double hilim = lookop_curve(conf->rolloff.high_satscale, hsv.v[0] / 6.0 * 360);
        double sat_scale = ahsv.v[1] <= 0.0 ? 1.0 : hsv.v[1] / ahsv.v[1]; // saturation scale compared to Adobe curve
        if (sat_scale < lolim) {
            keep = 1.0;
        } else if (sat_scale < hilim) {
            // S-curve transition
            double x = (sat_scale - lolim) / (hilim - lolim); // x = [0..1], 0 at lolim, 1 at hilim
            if (x < 0.5) {
                x = 0.5 * pow(2*x, 2);
            } else {
                x = 0.5 + 0.5 * (1-pow(1-2*(x-0.5), 2));
            }
            keep = 1.0*(1.0-x) + keep*x;
        } else {
            // do nothing, very high increase, keep minimum amount
        }
        if (newLuminance < oldLuminance) { // only do this for colors that are brightened
            double ls = srgb_gamma_forward(newLuminance) / srgb_gamma_forward(oldLuminance);
            const double lolim = 0.75;
            const double hilim = 1.0;
            if (ls < lolim) {
                keep = 1.0;
            } if (ls < hilim) {
                // S-curve transition
                double x = (ls - lolim) / (hilim - lolim); // x = [0..1], 0 at lolim, 1 at hilim
                if (x < 0.5) {
                    x = 0.5 * pow(2*x, 2);
                } else {
                    x = 0.5 + 0.5 * (1-pow(1-2*(x-0.5), 2));
                }
                keep = 1.0*(1.0-x) + keep*x;
            }
        }
        if (keep < 1.0) {
            // mix in some of the Adobe curve result, but we hue-correct it first
            xyz = m3x3_mul_v3(rgb2xyz, argb);
            XYZ = (cmsCIEXYZ){ .X = xyz.v[0] * 100.0, .Y = xyz.v[1] * 100.0, .Z = xyz.v[2] * 100.0 };
            cmsCIECAM02Forward(state->h02, &XYZ, &JCh);
            JCh = (cmsJCh){ JCh.J, JCh.C, jch.v[2] };
            cmsCIECAM02Reverse(state->h02, &JCh, &XYZ);
            xyz = (v3){{ XYZ.X / 100.0, XYZ.Y / 100.0, XYZ.Z / 100.0 }};
            v3 cc_argb = v3_clip01(m3x3_mul_v3(xyz2rgb, xyz));
            if (!isfinite(cc_argb.v[0]) || !isfinite(cc_argb.v[1]) || !isfinite(cc_argb.v[2])) {
                cc_argb = argb;
            }
            for (int i = 0; i < 3; i++) {
                rgb.v[i] = srgb_gamma_forward(rgb.v[i]) * keep + srgb_gamma_forward(cc_argb.v[i]) * (1.0 - keep);
                rgb.v[i] = srgb_gamma_inverse(rgb.v[i]);
            }
        }
    }

    if (state->conf.lookop_count > 0) {
        xyz = m3x3_mul_v3(rgb2xyz, rgb);
        XYZ = (cmsCIEXYZ){ .X = xyz.v[0] * 100.0, .Y = xyz.v[1] * 100.0, .Z = xyz.v[2] * 100.0 };
        cmsCIECAM02Forward(state->h02, &XYZ, &JCh);
        if (!isfinite(JCh.J) || !isfinite(JCh.C) || !isfinite(JCh.h)) {
            return false;
        }

        jch = apply_look_operators((v3){{ JCh.J, JCh.C, JCh.h }}, rgb2xyz, xyz2rgb, state);
        if (!isfinite(jch.v[0]) || !isfinite(jch.v[1]) || !isfinite(jch.v[2])) {
            return false;
        }

        jch = desaturate_to_fit(jch, xyz2rgb, state->h02);

        JCh = (cmsJCh){ jch.v[0], jch.v[1], jch.v[2] };
        cmsCIECAM02Reverse(state->h02, &JCh, &XYZ);
        xyz = (v3){{ XYZ.X / 100.0, XYZ.Y / 100.0, XYZ.Z / 100.0 }};
        rgb = m3x3_mul_v3(xyz2rgb, xyz);
        for (int i = 0; i < 3; i++) {
            if (!isfinite(rgb.v[i])) return false;
            if (rgb.v[i] < 0) rgb.v[i] = 0;
            if (rgb.v[i] > 1) rgb.v[i] = 1;
        }
    }

    *out = m3x3_mul_v3(rgb2xyz, rgb);
    return true;
}
